#include "gpl.h"
#include "stdafx.h"
#define STAND_ALONE 0
#if !STAND_ALONE
#include "DePbo.h"
#else
#define DEPBODLL_API
#include "Bin2CppClass.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdarg.h>
#include <io.h>
#include "PboUtils.h"

#define USE_QUOTES	0

////////////////////////////
bincpp::bincpp()
{
	StringTable=NULL;
	InBuffer=NULL;
	OutBuffer=NULL;
	from_file=false;
	Close();
}
bincpp::~bincpp()
{
	Close();
}

// used to release buffer to caller, which then becoems caller's responsibiltiy

PUBLIC void bincpp::Nullify()
{
	OutBuffer=NULL;
}

PUBLIC void bincpp::Close()
{
	if (OutBuffer) free(OutBuffer);
	OutBuffer=NULL;
	if (from_file && InBuffer) free(InBuffer);
	InBuffer=NULL;
	from_file=false;

struct 	StringTable *st=StringTable,*old;

	while (st)
	{
		old=st;
		st=st->next;
		free(old);
	}
	StringTable=NULL;
	OutCount=OutSize=
	bufsize=InCount=0;
}
#define HEADER_SIZE 7
static char header[HEADER_SIZE]={0,'r','a','P',4,0,0};
PUBLIC int bincpp::Open(const char *fname)
{
long hFile;
struct _finddata_t c_file;
FILE *fpin;

	Close();
	from_file=false;
    if( (hFile = _findfirst( fname, &c_file )) == -1L ) return 0;
	_findclose(hFile);
	if (c_file.size<=HEADER_SIZE) return 0;
	if (!(InBuffer=(uchar *)malloc(c_file.size))) return 0;
	if (!(fpin=fopen(fname,"rb"))) goto err;
	bufsize=fread(InBuffer,1,c_file.size,fpin);
	fclose(fpin);
	if (c_file.size!=bufsize)
	{
err:
		free(InBuffer);
		InBuffer=NULL;
		return 0;
	}
	if (strncmp((char *)InBuffer,"\000raP\004\000",HEADER_SIZE)) goto err;

	InCount=HEADER_SIZE;
	from_file=true;
	return  bufsize;
}
PUBLIC void bincpp::SetData(uchar *buf,unsigned len)
{
	Close();
	InBuffer=buf;
	bufsize=len;
	InCount=7;
}
PUBLIC bool bincpp::Write(const char *fname,const char *buffer)
{
FILE *fp=fopen(fname,"wb");
unsigned len=strlen(buffer);
bool sts;
	if (!fp) return false;
	sts=fwrite(buffer,1,len,fp)==len;
	fclose(fp);
	return sts;
}

/////////////////////////
// recursive routine
/////////////////////////
PUBLIC char * bincpp::ReadPackets() 
{
	if (!OutputStr(0, "//extracted with bincpp.dll\r\n")) return NULL;
	if (ReadPacket() && ReadDefines() )// if any
			return OutBuffer;
	if (OutBuffer) free (OutBuffer);
	return OutBuffer=NULL;
}

PRIVATE bool bincpp::ReadPacket(int indent) 
{
	switch(GetByte()) // type
    {
	default:												break;
    case PKT_CLASS:		if (ReadClassPacket(indent))		return true; break;
    case PKT_VARIABLES:	if (ReadVariablesPacket(indent))	return true; break;
    case PKT_ARRAYS:	if (ReadArrayPacket(indent))		return true; break;
    }
	return false;
}
///////////////////////////////////////////
// class ClassName : InheritedClass { ....}
////////////////////////////////////////////
PRIVATE bool bincpp::ReadClassPacket(int indent)
{
const char *name,*inheritclass;
int children;

	if (!OutputStr(0, "\r\n")) return false;
    if (!(name=ReadIndexedStr())) return false;
	if (!(inheritclass=ReadStr())) return false;
	if (*inheritclass)
	{
		if (!OutputStr(indent, "class %s: %s {\r\n", name, inheritclass)) return false;
	}
	else if (!OutputStr(indent, "class %s {\r\n", name)) return false;
    if (EOF==(children=ReadShort())) return false;
	while (children--) if (!ReadPacket( indent + 1)) return false;
    return OutputStr(indent, "};\r\n\r\n");
}
//////////////////////////////////////
// name="string";
// name=1.23;
// name=123;
// name[]={...}
//////////////////////////////////////
bool bincpp::ReadVariablesPacket(int indent)
{
int VarType;
const char *name;

	if (EOF==(VarType = GetByte())) return false;
    if (!(name=ReadIndexedStr())) return false;
	if (!format(VarType,0,indent,name)) return false;
	return true;
}
bool bincpp::ReadArrayPacket(int indent)
{
const char *name;

	if (!(name=ReadIndexedStr())) return false;
	if (!OutputStr(indent, "%s[] = { ", name)) return false;
	if (!read_array(indent + 1)) return false;
	return OutputStr(0, "};\r\n");
}
/// recursive
bool bincpp::read_array(int indent)
{
int nTypes;
   
    if (EOF==(nTypes=ReadShort())) return false;
    
    while (nTypes--)
		if (!format(GetByte(),nTypes,indent))
			return false;
	return true;
}

bool bincpp::format(int VarType,int sz,int indent,const char *name)
{
bool sts;

	if (VarType==EOF) return false;
	switch(VarType)
    {
    default: return false;
    case VAR_STRING:	
	{
	const char *n2,*ptr;

		if (!(n2=ReadIndexedStr())) return false;
#if !USE_QUOTES
		/* what's important here is multiple commands on same line etc
		such as init strincgs
		binarized files dont have begiining and ending quotes per se

		problems with pre-existing quotes
		*/
		// first, establish if quotes are needed
//		if (*n2=='"' && n2[strlen(n2)-1]=='"') goto noq;
		if (!*n2)
		{
			if (name)	sts=OutputStr(indent, "%s = \"\";\r\n", name);
			else		sts=OutputStr(0, "\"\"");
			break;
		}
		
		for (ptr=n2;*ptr;ptr++)
		{
			switch(*ptr)
			{
			case ':':
			case ',':
			case '=':
			case ';':
			case '"':
			case '{':
			case '(':
			case '}':
			case ')':
			case '[':
			case ']': goto needs_quotes;
			}
		}
		if (name)	sts=OutputStr(indent, "%s = %s;\r\n", name, n2);
		else		sts=OutputStr(0, "%s", n2);
		break;
needs_quotes:
	
// otherwise we need to strudp
		char *nv=(char *)malloc(2*strlen(n2));
		if (!nv) return false;
		char *optr=nv;
		for (ptr=n2;*ptr;ptr++)
		{
			if ((*optr++=*ptr)=='"') *optr++='"';
		}
		*optr=0;
		if (name)	sts=OutputStr(indent, "%s = \"%s\";\r\n", name, nv);
		else		sts=OutputStr(0, "\"%s\"", nv);
		free(nv);
		break;
#else
	// problems here if we have 'jehlan cs (.p3d)
		if (name)	sts=OutputStr(indent, "%s = \"%s\";\r\n", name, n2);
		else		sts=OutputStr(0, "\"%s\"", n2);
        break;
#endif
	}
	
	case VAR_FLOAT:
    {
	float fl = GetFloat();
	
		if (name)   sts=OutputStr(indent, "%s = %f;\r\n", name, fl);
		else		sts=OutputStr(0, "%f", fl);
        break;
    }

    case VAR_INTEGER:
	{
	int val=GetLong();
    
		if (name)   sts=OutputStr(indent, "%s = %d;\r\n", name, val);
        else		sts=OutputStr(0, "%d", val);
        break;
	}
    
	case VAR_ARRAY:
        
		if (name) return false;
		if (!OutputStr(0, "\r\n")) return false;
        if (!OutputStr(indent, "{ ")) return false;
        if (!read_array(indent + 1)) return false;
        sts=OutputStr(0, " }");
        break;
    }
	if (!sts) return false;
    if (sz) if (!OutputStr(0, ", ")) return false;

    if (VarType == 3)
    {
       if (!OutputStr(0, "\r\n")) return false;
       if (!OutputStr(indent, "")) return false;
    }
	return true;
}

/*
 * Reads a string id from the file stream, if it's unknown read
 * following string also.
 */

const char *bincpp::ReadIndexedStr()
{
int id;
const char *ptr;
struct 	StringTable *st=	StringTable;


    if (EOF==(id=ReadShort())) return NULL;

	while (st)
	{
		if (st->id==id) return st->str;
		st=st->next;
	}

    if (!(ptr=ReadStr())) return NULL;
	struct 	StringTable *stt =new 	struct StringTable;
	stt->next=NULL;
	stt->id=id;
	stt->str= ptr;
	if (!(st=StringTable)) 	StringTable=stt;
	else 
	{
		while (st->next) st=st->next;
		st->next=stt;
	}
	return ptr;
}
/*
 * Read a string without stringtable magic.
 */

const char *bincpp::ReadStr()
{
const char *ptr;
int len;
	
	if (InCount>=bufsize) return NULL;
	len=strlen((char *)&InBuffer[InCount])+1;
	if (len+InCount>=bufsize) return NULL;
	ptr=(char *)&InBuffer[InCount];
	InCount+=len;
	return ptr;
}


/*
 * This is a compressed 2 byte number for ID's and Sizes
 * Am using -1 as an impossible value
 */

int bincpp::ReadShort()
{
int val;

    if ((val = GetByte())==EOF) return EOF;
    if (val & 0x80)
    {
		int extra;
        if ((extra = GetByte())==EOF) return EOF;
        val += (extra - 1) * 0x80;
    }
	return val;
}

float bincpp::GetFloat()
{
union
{
	float f;
	char a[4];
}val;

	val.a[0]=GetByte();
	val.a[1]=GetByte();
	val.a[2]=GetByte();
	val.a[3]=GetByte();
	return val.f;
}

int bincpp::GetLong()
{
union
{
	int f;
	char a[4];
}val;

	val.a[0]=GetByte();
	val.a[1]=GetByte();
	val.a[2]=GetByte();
	val.a[3]=GetByte();
	return val.f;
}
int bincpp::GetByte()
{
	if (InCount>=bufsize)		return EOF;
	return (unsigned char)(InBuffer[InCount++]);
}
bool bincpp::ReadDefines()
{
	if (InCount>=bufsize)		return true; // none
	if (InCount+4 >bufsize) return false; // huh?

long count=GetLong();
const char *name;
int val;

	while (count--)
	{
		if (!(name=ReadStr())) return false;
		val=GetLong();
        OutputStr(0, "#define %s\t%d\r\n", name, val);
    }
	return true;
}
PRIVATE bool bincpp::AddToOutput(char *buf)
{
int len=strlen(buf);
	if (!len) return true;
	if ((OutCount+len) >= OutSize)
	{
		OutSize=2* OutSize+len;
		OutBuffer=(char *)realloc(OutBuffer,OutSize);
		if (!OutBuffer)
		{
			OutSize=OutCount=0;
			return false;
		}
	}
	strcpy(&OutBuffer[OutCount],buf);
	OutCount+=len;
	return true;
}

PRIVATE bool bincpp::OutputStr(int indent, char *fmt, ...)
{
char buf[10240]; // any more than that = too bad

    if ((indent < 0)  && !AddToOutput("//") ) return false;
    
    if (indent>0)
	{
		buf[0]=0;
		while (indent--) strcat(buf,"\t");
		if (!AddToOutput(buf)) return false;
	}    
    va_list ap;    
    va_start(ap, fmt);    
	vsprintf(buf,fmt,ap);
    va_end(ap);
	return AddToOutput(buf);
}



