#ifndef BIN2CPPh
#define BIN2CPPh

#include "../gpl.h"			// license agreement
#include "../stdtyp.h"
#include <stdio.h>


// this is a standalone class, incorporated in the container dll
// that is used internally by the container if it detects magic signatures
// or can be called, uniquely, without reference to a pbo

#define PACKTYPES_DEFINED	1

enum packet_types
{
	PKT_CLASS,
	PKT_VARIABLES,
	PKT_ARRAYS
};
enum var_types
{
	VAR_STRING,
	VAR_FLOAT,
	VAR_INTEGER,
	VAR_ARRAY
};

class DEPBODLL_API bincpp
{
public: 
	bincpp();
	~bincpp();
	int	Open(const char *fname);			// Either open a file
	void SetData(uchar *data,unsigned len); // OR, pass a data reference
											// both of these cause previous stuff to be closed
	void Close();							// and dispose of either
	char * ReadPackets();					// interpret the data, or file
	void Nullify();				// caller takes responsibility for freeing out buffer
// be careful to Close a file above rather than rely on open to close a file
// if you use nullify()
	bool		Write(const char *fname,const char *buffer);
private:
	bool ReadDefines();
	bool ReadPacket(int indent=-1);
	bool OutputStr(int indent, char *fmt, ...);
	bool AddToOutput(char *buf);
	bool read_array(int indent);
	bool ReadVariablesPacket(int indent);
	bool ReadArrayPacket(int indent);
	bool ReadClassPacket(int indent);
	float GetFloat();
	int GetLong();
	int GetByte();
	int ReadShort();
	const char *ReadStr();
	const char *ReadIndexedStr();

	struct StringTable
	{
		int	id;
		const char	*str;
		StringTable *next;
	}*StringTable;
	uchar *InBuffer;
	char *OutBuffer;
	unsigned bufsize,InCount;
	bool format(int tmp,int sz,int indent,const char *name =NULL);
	unsigned OutCount,OutSize;
	bool from_file;
};

#endif
