#include "..\gpl.h"
#include "stdafx.h"
#include "DePbo.h"
#include "ErrTypes.h"
#include "Reader.h"
#include "Writer.h"
#include "..\stdtyp.h"
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#define TestForEol() (Read.GetToken(GET_TOKEN_TEST_EOL)==';')?NO_ERROR:ERR_SEMICOLON_EXPECTED

// coc files sometimes have ; as the comma separator in unit[]= arrays

LOCAL bool AllowSemiAsComma=false;
LOCAL bool IncludeDetected=false;// used to prevent a write on what would be garbage translation


PUBLIC class DefList *DefList;
/////////////////////////////////////////////////////////
int	Cpp2BinClass::GetReturnedToken()
{
	return ::Read.GetTokenReturn();
}

const char *Cpp2BinClass::ErrString(int code)
{
	switch(code)
	{
	default: return "unknown error type";
	case NO_ERROR: return "No Error";
	case ERR_NO_MEMORY: return "No Memory";
	case ERROUT_UNKNOWN_PACKET_TYPE: return "Output Unknown packet type decoded (internal error)";
	case ERROUT_UNKNOWN_ARRAY_TYPE: return "Internal error: unknown array type decoded";
	case ERR_OUTPUT: return "Cannot write to file";
	case ERR_UNEXPECTED_BRACE: return "Unexpected {";
	case ERR_SEMICOLON_EXPECTED: return "Expected Semicolon (or eol)";
	case TOKEN_EOF:	return "EOF encountered";
	case TOKEN_ERR_MISSING_QUOTE: return "Mismatched quotes";
	case TOKEN_ERR_NO_TOKEN: return "No Read.Token";
	case ERR_TOKEN_EXTRANEOUS1: return "Expected class, define, or varname";
	case ERR_EXPECTED_CLASSNAME: return "Expected class name";
	case ERR_EXPECTED_INHERIT_OR_BRACE: return "Expected Inherit or {";
	case ERR_EXPECTED_INHERIT_LABEL: return "Expected Inheritance label";
	case ERR_EXPECTED_BRACE_AFTER_INHERIT: return "Expected Brace after inherit";
	case ERR_EXPECTED_ARRAY_OR_EQUALS: return "Expected array[] or =";
	case ERR_EXPECTED_EQUALS: return "Expected =";
	case ERR_EXPECTED_BRACE_ON_ARRAY: return "Expected { on array";
	case ERR_EXPECTED_BRACKET: return "Expected ]";
	case ERR_EXPECTED_VALUE: return "Expected value after =";
	case ERR_ARRAY_EXPECTED_VALUE: return "Expected value , or {} in array";
	case ERR_DEFINE_BAD_LABEL: return "#Define bad label";
	case ERR_DEFINE_BAD_VALUE: return "#define no value";
	case ERR_DEFINE_NOT_INTEGER: return "#define not integer";
	case ERR_CLOSEPAR_EXPECTED: return "Expected ) during define expansion";
	case ERR_NO_DEFINE: return "No define() found";
	case ERR_DEFINE_PARMCOUNT: return "Not enough parms in define expansion";
	case ERR_NO_PARMEND: return "No end ) on definition";
	case TOKEN_EOL: return "no value for define"; // nornally this is ok
	case ERR_NO_FILE: return "cannot open file";
	case ERR_HASH_DET: return "Cannot Handle ## in defines";
	case ERR_INCLUDE_DET: return "Cannot write an include file";
	}
}
//////////////////////////////////////////////////////////////
Cpp2BinClass::Cpp2BinClass()
{
	First=NULL;
}

Cpp2BinClass::~Cpp2BinClass()
{
	Clear();
}

PRIVATE void Cpp2BinClass::Clear()
{
	if (First) delete First;
	First=NULL;
	IncludeDetected=false;
}
///////////////////////////////////////////
// the beginning of it all
// we instantiate Writer and deflist during function calls
// to cut down on dll memory use if it's never called
//////////////////////////////////////////////////////

PUBLIC int Cpp2BinClass::ProcessFile(const char *fname)
{

	if (!::Read.OpenFile(fname)) return ERR_NO_FILE;
	return Start(fname);
}

PUBLIC int Cpp2BinClass::ProcessBuffer(const char *BufferName,const char *fname,unsigned sz)
{
	::Read.SetBuffer(BufferName,sz);
	return Start(fname);
}

PRIVATE int Cpp2BinClass::Start(const char *fname)
{
int sts=ERR_NO_MEMORY;
	Clear();
	if (!(DefList = new class DefList())) goto bad;
	if (!(First=new class Classes())) goto bad;
	if (!(First->Name=strdup(fname))) goto bad;
	First->Type=VARTYPE_CLASS;
	sts= First->ProcessClass(true);
bad:
	if (sts) Clear();
	// otherwise, the values decoded will be used by someone
	if (DefList) delete DefList;
	DefList=NULL;
	::Read.Close();
	return sts;
}

//////////////////////////////////////////////////////////////////
PUBLIC int Cpp2BinClass::Output(const char *filename)
{
int sts=NO_ERROR;
class Writer Write;
	if (IncludeDetected) return ERR_INCLUDE_DET;
	if (!Write.Open(filename)) return ERR_OUTPUT;
	if ((sts=First->Output(&Write))) 	return sts;
	if (!Write.Long(0)) return ERR_OUTPUT;// there is no integer deflist
	return sts;
}
PUBLIC Classes *Cpp2BinClass::FindClass(const char *cname)
{
Classes *cl=First;// is th ename of the file
/// it only has a single entry, the embedded list of classes and vars
// there is no 'next' for First

	for (cl=cl->embedded;cl;cl=cl->next)
	{
		if (cl->Type==VARTYPE_CLASS && !_stricmp(cl->Name,cname))
			return cl;
	}
	return NULL;
}

PRIVATE int Classes::Output(class Writer *Write)
{
int sts;

	switch(Type) // whoami
	{
	default: return ERROUT_UNKNOWN_PACKET_TYPE;
	case VARTYPE_CLASS:

		if (!Write->ClassName(this,Count())) return ERR_OUTPUT;
		if (embedded && ((sts=embedded->Output(Write)))) return sts;
		break;

	case VARTYPE_FLOAT:
	case VARTYPE_INTEGER:
	case VARTYPE_STRING:

		if (!Write->Variable(this)) return ERR_OUTPUT;
		break;
	
	case VARTYPE_ARRAY:

		if (!Write->ArrayName(Name,Count())) return ERR_OUTPUT;
		if (embedded && ((sts=embedded->OutputArray(Write)))) return sts;
		break;
	}
	if (next) return next->Output(Write);

	return NO_ERROR;
}

int Classes::OutputArray(class Writer *Write)
{
int sts;

	switch(Type)
	{
	default: return ERROUT_UNKNOWN_ARRAY_TYPE;
	case VARTYPE_INTEGER:
	case VARTYPE_FLOAT:
	case VARTYPE_STRING:
	
		if (!Write->ArrayVariable(this)) return ERR_OUTPUT;
		break;
	
	case VARTYPE_ARRAY:

		if (!Write->ArraySpecial(Count())) return ERR_OUTPUT;
		if (embedded && ((sts=embedded->OutputArray(Write)))) return sts;
		break;
	}
	if (next) return next->OutputArray(Write);
	return NO_ERROR;
}

PRIVATE int Classes::Count()
{
Classes *cl=embedded;

	if (!cl) return 0;

int cnt=1;

	while (cl->next){cl=cl->next; cnt++;}

	return cnt;
}
//////////////////////////////////////////////////
Classes::Classes()
{
	Name=Value=NULL;
	embedded=next=NULL;
	Type=VARTYPE_NONE;
}
Classes::~Classes()
{
	if (Name)	free (Name);
	if (Value) free(Value);
	if (next) delete next;
	if (embedded) delete embedded;
}


PRIVATE Classes *Classes::NewClass()
{
Classes *clnew;

	clnew=new Classes();
	if (!embedded) return embedded=clnew;

Classes *cln=embedded;
	while (cln->next) cln=cln->next;
	return cln->next=clnew;
}
PUBLIC int Classes::ProcessClass(bool start)
{
int sts;

	for(;;)
	{
		switch(Read.GetToken())
		{
		case TOKEN_EOF: if (start) return NO_ERROR;
		default: return ERR_TOKEN_EXTRANEOUS1;
		case '}': 

			if (start) return ERR_UNEXPECTED_BRACE;
			return TestForEol();

		case TOKEN_VALID:

			if (!stricmp(Read.Token,"#define")) 
			{
				if (Read.GetToken(GET_TOKEN_HUNT_DEF_LABEL)!=TOKEN_VALID) return ERR_DEFINE_BAD_LABEL;
				char *label=Read.DupToken();
				if (!label) return ERR_NO_MEMORY;
				if ((sts=Read.GetToken(GET_TOKEN_HUNT_EOL))==TOKEN_EOL) return sts;
				sts=	DefList->Add(label,Read.Token);
				free(label);
			}
			else
			{
				if (!stricmp(Read.Token,"#include"))
				{
					if ((sts=Read.GetToken(GET_TOKEN_HUNT_EOL))==TOKEN_EOL) return sts;
					IncludeDetected=true;

				}
				else
				{
			// create the embedded ptr, or, add to existing
					Classes *clnew;

					if (!(clnew=NewClass())) return  ERR_NO_MEMORY;
					if (!stricmp(Read.Token,"class")) sts=clnew->AddClass();
					else						 sts=clnew->AddVariable();
				}
			}
			if (sts) return sts;
			printf("*");

		}
	}
}
// class thing:inherit {.....};
PRIVATE int Classes::AddClass()
{
	Type=VARTYPE_CLASS;
	if (Read.GetToken()!=TOKEN_VALID) return ERR_EXPECTED_CLASSNAME;
	if (!(Name=Read.DupToken())) return ERR_NO_MEMORY;
	// there may be nothing to embed, let alone add

	switch(Read.GetToken())
	{
	default: return ERR_EXPECTED_INHERIT_OR_BRACE;
	case ':':

		if (Read.GetToken()!=TOKEN_VALID) return ERR_EXPECTED_INHERIT_LABEL;
		if (!(Value=Read.DupToken())) return ERR_NO_MEMORY;
		if (Read.GetToken()!='{') return ERR_EXPECTED_BRACE_AFTER_INHERIT;

	case '{': return ProcessClass();
	}
}

	// array	weapons[] = {};
	// float	requiredVersion = 1.75;
	// integer	side=0;
	// string	displayName="Mechanic";

int Classes::AddVariable()
{
	if (!(Name=Read.DupToken())) return ERR_NO_MEMORY;
	switch(Read.GetToken())
	{
	default: return ERR_EXPECTED_ARRAY_OR_EQUALS;
	case '[': 
		
		if (Read.GetToken()!=']') return ERR_EXPECTED_BRACKET;
		if (Read.GetToken()!='=') return ERR_EXPECTED_EQUALS;
		if (Read.GetToken()!='{') return ERR_EXPECTED_BRACE_ON_ARRAY;

		return AddArray();

	case '=': break;
	}
	if (!strcmp("init",Name))
	{
		int z=1;
	}
	switch(Read.GetToken(GET_TOKEN_ALLOW_SPACES))
	{
	default:	return ERR_EXPECTED_VALUE;
	case ';':	return AddValue("")?NO_ERROR: ERR_NO_MEMORY;

	case TOKEN_VALID: break;
	}
	if (!AddValue(Read.Token)) return ERR_NO_MEMORY;
	return TestForEol();
}

// a,b,c,d};
// {a,b},{c,d}};
int Classes::AddArray(bool embed)
{
int sts;
Classes *cl;
	// embedded
// no name or value is associated with embedded arrays
	// the class containing an embedded array has an embedded pointer, and sometimes a next
	Type=VARTYPE_ARRAY;	// no name associated
						//or  name already stored
	// there may not be any
bool has_value;
// this routine will miss
// {value ,} // it will not detect the comma, as an error
	for(;;)
	{
		switch(Read.GetToken())
		{
		default: return ERR_ARRAY_EXPECTED_VALUE;
		case '}':if (embed) return NO_ERROR;
				 return TestForEol();
		case '{':

			has_value=true;
			if (!(cl=NewClass())) return ERR_NO_MEMORY;
			if ((sts= cl->AddArray(true))) return sts;
			break;
	
		case TOKEN_VALID: 

			has_value=true;
			if (!(cl=NewClass())) return ERR_NO_MEMORY;
			if (!cl->AddValue(Read.Token)) return ERR_NO_MEMORY;
			break;

		case ';': if (!AllowSemiAsComma) return ERR_ARRAY_EXPECTED_VALUE;
		case ',': 
			if (!has_value) 
			{
				return ERR_ARRAY_EXPECTED_VALUE;
			}
			has_value=false;
			break;
		}
	}
}

int	Classes::AddValue(const char *tkn)
{
	if (!(Value=strdup(tkn))) return false;
	Type=VARTYPE_STRING; // assume
	if (!isdigit(*tkn) && *tkn!='.' && *tkn!='-') return true;

bool floater=false;

	if (*tkn=='-')
	{
		tkn++;
		if (!*tkn) return true;
	}

	for  (;*tkn;tkn++)
	{
		if (!floater && *tkn=='.') {floater=true; continue;}
		if (!isdigit(*tkn)) return true;
	}
	Type= floater?VARTYPE_FLOAT:VARTYPE_INTEGER;
	return true;
}
