#ifndef DEPBOh
#define DEPBOh

#include "../gpl.h"		// license agreement

#include <stdio.h>

// The following ifdef block is the standard way of creating macros which make exporting 
// from a DLL simpler. All files within this DLL are compiled with the PBODLL_EXPORTS
// symbol defined on the command line. this symbol should not be defined on any project
// that uses this DLL. This way any other project whose source files include this file see 
// PBODLL_API functions as being imported from a DLL, wheras this DLL sees symbols
// defined with this macro as being exported.

#ifdef DEPBO_EXPORTS
#define DEPBODLL_API __declspec(dllexport)
#else
#define DEPBODLL_API __declspec(dllimport)
#endif

#include "Bin2CppClass.h"
#include "Cpp2BinClass.h"

#define uchar unsigned char

#define DEPBO_VERSION "1.11"
// 1.1 fixed too small stack on bin2cpp
enum
{
	PBO_STS_NO_ERROR=0,
	PBO_STS_NO_MEMORY=-1,
	PBO_STS_FILE_TRUNCATED=-2,
	PBO_STS_NO_FILE=-3,
	PBO_STS_OUT_OF_BOUNDS=-4,
	PBO_STS_CHUNK_ZERO=-5,
	PBO_STS_CHECKSUM=-6,
	PBO_STS_NO_ENTRY=-7,
	PBO_STS_MULIPLE_PRODUCTS=-8,
	PBO_STS_UNKNOWN_COMPRESSION=-9
	,PBO_STS_NON_ZERO_FIELD=-10
	,PBO_STS_RESERVED_NONZERO=-11
};

typedef int pbo_status;

struct pbo_header
{
	char    *filename;	// 0 terminated
    unsigned type;		// 0 not packed, 0x43707273 - packed and sometimes another val
    unsigned real_size;	// 0 not packed or size of unpacked (original) file;
						// or even same same value as 'size'
    unsigned reserved;	// always 0
    unsigned time;		// stadard udt in seconds
    unsigned size;		// raw data. Same for packed and unpacked files. 
};

class pbo_entry 
{
private:					// only allow the container access into here

	friend class pboContainer;

	pbo_entry(FILE *fpPbo=NULL);			// reads file, NULL = initialise only
	~pbo_entry();
	pbo_status		 ReadHeader(FILE *fpPbo);	// if not from above
	pbo_status		 ReadRawData(FILE *fpPbo, long biased_offset);
	pbo_entry		*next;					// pointer to next one
	struct	pbo_header	hdr;
	uchar			*data;					// where the data is
	unsigned		 actual_size;
	long			 offset;				// how far into file after header
	bool			 compressed_file;
	bool			 extracted_file;
	pbo_status		 status;	
};

class  DEPBODLL_API pboContainer 
{
public:
    pboContainer();
	~pboContainer();

	pbo_status	Open(const char *pbo_filename);
	void		Close();
	pbo_entry	*FindEntry(char *fname);					// or null
	pbo_entry  *GetNextEntry(pbo_entry *e){return e->next;}	// iterate
	int			GetData(pbo_entry *entry);	// or returns -status
	int			GetData(char *fname);		// ditto
	const uchar *GetDataPtr(pbo_entry *e){	return e->data;}// after GetData
	void		FreeData(pbo_entry * entry);// releases buffer from above
	int		IsCompressed(pbo_entry * entry);
	int		WasEncrypted(pbo_entry * entry);
	const char *GetEntryFilename(pbo_entry *e){	return e->hdr.filename;}
	const char *strError(pbo_status pboStatus);
	const char *strError();
	int			GetEntryCount() { return file_count;}
	pbo_entry * FirstEntry;
// below is also used internally if mission.sqm (or any other 'file') contains
// magic signatures
// these 3 functions can be used as standalone for cpp2bin applications
	int			OpenBinCpp(char *fname){return bincpp.Open(fname);}
	char *		ReadBinCpp() {return bincpp.ReadPackets();}
	void		CloseBinCpp() {bincpp.Close();};
	bincpp			bincpp;	// same functions can be accessed directly
	Cpp2BinClass	cppbin;
	static const char* Version;
	struct
	{
		char *	Type;
		char *	Name;
		char *	Extension;
	}Product;
	int			strange_values; // 1 = timestamp = -1 on blank end
private:
 	unsigned	file_count;
	long		file_size;
    long		real_file_size;
	int			ReadHeaders();
	long		data_offset;
	FILE		*fpin;
	void		Init();
	pbo_status	status;
};


#endif
