#include "gpl.h"
#include "stdafx.h"
#include "DePbo.h"
#include <string.h>
#include <malloc.h>

#define PRIVATE
#define PUBLIC

#define uchar unsigned char


enum compression_types
{
	ENTRYTYPE_NORMAL,
	ENTRYTYPE_PRODUCT=0x56657273,	//1449489011 decimal
	ENTRYTYPE_COMPRESSED=0x43707273	// decimal 1131442803 (standard sig)
};

static char * read_string(FILE *fpPbo);

static pbo_status unpack_data(uchar *CompressedBuffer, size_t len, uchar *to,size_t DecompressedBufferSize);
const char *pboContainer::Version=DEPBO_VERSION;

#define AVOID_STACK_ERRORS	1


PUBLIC pbo_entry::pbo_entry(FILE *fpPbo)
{
	hdr.filename=NULL;
	hdr.type=hdr.real_size=hdr.reserved=hdr.time=hdr.size=0;
	actual_size=0;
	compressed_file=extracted_file=false;
	next=NULL;
    data = NULL;
	if (fpPbo) status=ReadHeader(fpPbo);
	else	status=PBO_STS_NO_FILE;
}

PUBLIC pbo_status pbo_entry::ReadHeader(FILE *fpPbo)
{
	if (!fpPbo) return status=PBO_STS_NO_FILE;

	status=PBO_STS_FILE_TRUNCATED; // assume
	if (!(hdr.filename=read_string(fpPbo))) return 	status;
    if (fread(&(hdr.type), 1, 4, fpPbo)!=4) return status;
    if (fread(&(hdr.real_size), 1, 4, fpPbo)!=4) return  status;
    if (fread(&(hdr.reserved), 1, 4, fpPbo)!=4) return status;
    if (fread(&(hdr.time), 1, 4, fpPbo)!=4) return status;
    if (fread(&(hdr.size), 1, 4, fpPbo)!=4) return status;
    if (!hdr.real_size || hdr.real_size==hdr.size) actual_size=hdr.size;
	else
	{
		compressed_file=true;
		actual_size=hdr.real_size;
	}
	return 	status=PBO_STS_NO_ERROR; 
}

pbo_entry::~pbo_entry()
{
	if (hdr.filename) 		free(hdr.filename);
    if (data)		free(data);
#if !AVOID_STACK_ERRORS
	if (next) delete (next);
#endif
}

static char * read_string(FILE *fpPbo)
{
int c;
int i=0;
char tmp[256];

	if (!fpPbo) return NULL;
    do
    {
		if ((c = fgetc(fpPbo))== EOF) return NULL;
        tmp[i++] = c;
		if (i==255)			return NULL; // rubbish
	}while(c);
 	return strdup(tmp);
}

/*
 * Read pbo_entry into *data
 */
int pbo_entry::ReadRawData(FILE *fpPbo, long data_offset)
{
	if (status) return status;
	if (data)  return actual_size;// already read

    if (fseek(fpPbo, data_offset + offset, SEEK_SET)) return status=PBO_STS_FILE_TRUNCATED;

    if (!(data = (uchar*)malloc(hdr.size))) return status=PBO_STS_NO_MEMORY;
    if (fread(data, 1,hdr.size,  fpPbo)!=hdr.size) return status=PBO_STS_FILE_TRUNCATED;
    if (!compressed_file) return actual_size;
    /* Data is packed */
uchar *space;


    if (!(space = (uchar*)malloc(hdr.real_size + 1))) return status=PBO_STS_NO_MEMORY;

	status=unpack_data(data, hdr.size, space,hdr.real_size);
	free(data);
    data= space;
    return  status?status:actual_size;
}

//////// this is THE decompression utility
/*
 * Unpacks data into the memory space pointed to by result
 */
static pbo_status 
unpack_data(uchar *CompressedBuffer, size_t PackedSize, uchar *DeCompressedBuffer,size_t DecompressedBufferSize)
{
uchar *CompressedPtr, *DecompressedPtr;
size_t offset;
int length, tmp, count;
unsigned check1, check2;
int rpos, rlen;
int chunk, size;
uchar index;

uchar b1, b2;

 	if ( (PackedSize-=4)<=0) return PBO_STS_OUT_OF_BOUNDS;
     size = 0;
	memset(DecompressedPtr =DeCompressedBuffer, 'A', DecompressedBufferSize);

    /*
     * Each data block is precded by a byte telling us what to do with
     * the next 8 bytes.
     */
	for (offset=0;offset < PackedSize; offset += length)
	{
        CompressedPtr = &CompressedBuffer[offset++];
        index = *CompressedPtr;
        length = 0;
        for (count = tmp=1;tmp < 256 && offset < PackedSize ; tmp*=2, count++)
        {
            CompressedPtr = &CompressedBuffer[offset];
            size = DecompressedPtr - DeCompressedBuffer;
          
            if (index & tmp)
            {
                /* Append the byte to output */
                *DecompressedPtr++ = *CompressedPtr;
                offset++;
            }
			else 
			{
                /* Read a pointer */
                b1 = *CompressedPtr;
                b2 = *(CompressedPtr + 1);
                offset += 2;

                rpos = size - b1 - 256 * ( b2/16 );
                rlen = b2 - 16 * (b2 / 16) + 3;
                // better with b2 & 0x0F + 3
                // printf("One annoying pointer (%d/%d/%d) (%d %d)\n", b1, b2,    size, rpos, rlen);

                if ((rpos + rlen) < 0)
                {
                    // printf("Padding With spaces\n");
					size+=rlen;
                    while (rlen--)   *DecompressedPtr++ = ' ';
                }
                
                /* This is not so nice */
                else 
					if ((DeCompressedBuffer + rpos) > DecompressedPtr)	return PBO_STS_OUT_OF_BOUNDS;
                    //printf("Size: %d Rpos: %d\n", size, rpos);
                /* PAD the file with a block from another place in the file */
	                else 
					if ((rpos + rlen) <= size)
		            {
		                /* I think we're supposed to add spaces then */
						while (rpos < 0)
						{
	                       *DecompressedPtr++ = ' ';
	                       rlen--;
	                       rpos++;                        
	                   }
                    
                    // printf("Easy append %d (%d)\n", rlen, rpos);
						memcpy(DecompressedPtr, &DeCompressedBuffer[rpos], rlen);
						DecompressedPtr += rlen;
						size += rlen;
	                }
              
                /* PAD the file with the block until size is rpos + rlen */
					else 
					if ((rpos + rlen) > size)
					{
						chunk = size - rpos;
						// printf("Size: %d Rpos: %d %d\n", size, rpos, chunk);
						while (rlen > 0)
						{
							if (chunk > rlen) chunk = rlen;
	                        if (!chunk) return PBO_STS_CHUNK_ZERO;
                        
                        // printf("Appended %db %d %d\n", chunk, size,
						    memcpy(DecompressedPtr, &DeCompressedBuffer[rpos], chunk);
							DecompressedPtr += chunk;
							size += chunk;
							rlen -= chunk;
						}
					} 
			}//!indextemp
        }//while temp
    }//while offset

    /* Last 4 bytes of the packed data is the checumsum, unsigned int */
    memcpy(&check1, &CompressedBuffer[PackedSize], sizeof(check1));
    
 	for (check2 = 0,DecompressedPtr = DeCompressedBuffer; DecompressedPtr < &DeCompressedBuffer[DecompressedBufferSize];) check2+= *DecompressedPtr++;

// the 'size' is often one less that true size, dunno why
    if (check1 != check2)			return PBO_STS_CHECKSUM;
	return PBO_STS_NO_ERROR;
}

///////////
PUBLIC pboContainer::pboContainer()
{
	Init();
}
PRIVATE void pboContainer::Init()
{
	fpin=NULL;
    FirstEntry = NULL;
    file_count = 0;
	Product.Type=Product.Name=Product.Extension=NULL;
    file_size = 0L;
    real_file_size = 0L;  
	data_offset=0;
	status=PBO_STS_NO_FILE;
}
PUBLIC pbo_status pboContainer::Open(const char *fname)
{
int sts;
	Close(); // if another was open
	if (!(fpin=fopen(fname,"rb"))) return status=PBO_STS_NO_FILE;
	if ((sts=ReadHeaders())) 
	{
		Close();
		status=sts;
	}
	return status;
}
PUBLIC void pboContainer::Close()
{
	if (fpin) fclose(fpin);
	if (Product.Name) free(Product.Name);
	if (Product.Extension) free(Product.Extension);
	if (Product.Type) free(Product.Type);

#if !AVOID_STACK_ERRORS
	if (FirstEntry) delete (FirstEntry);
#else
pbo_entry *n,*e=FirstEntry;

	while (e)
	{
		n=e->next;
		delete(e);
		e=n;
	}
#endif
	Init();
}
PUBLIC pboContainer::~pboContainer()
{
	Close();
}


PRIVATE pbo_status pboContainer::ReadHeaders()
{
pbo_entry *entry,*tail=NULL;

	strange_values=0;
    while (1)
    {
 // sub relies on caller to delete the chain 
		if (!(entry = new pbo_entry())) return PBO_STS_NO_MEMORY;

		if ((status=entry->ReadHeader(fpin))!=PBO_STS_NO_ERROR)
		{
			delete (entry);
			return status;
		}

        if (!strcmp(entry->hdr.filename, ""))
        {
			if (entry->hdr.real_size || entry->hdr.reserved || entry->hdr.size)
			{
				if (entry->hdr.time=0xffffffff)
				{
					strange_values=1;
					goto ignore;
				}
				delete(entry);
				return PBO_STS_NON_ZERO_FIELD;
			}
ignore:
			switch(entry->hdr.type)
			{
			case ENTRYTYPE_PRODUCT:
// this value is only known to occur as first entry
				if (Product.Type) return status=PBO_STS_MULIPLE_PRODUCTS;
				if (!(Product.Type=read_string(fpin))) return status; // "Product"
				if (!(Product.Name=read_string(fpin))) return status; // "OFP: Resistance"
				if (!(Product.Extension=read_string(fpin))) return status; // ""
				delete (entry);
				continue;
			
			case ENTRYTYPE_COMPRESSED:
			case ENTRYTYPE_NORMAL: // standard end

    /* Get the position, need it to read files later. */
			    data_offset = ftell(fpin);
	            delete (entry);
			    return PBO_STS_NO_ERROR;
			
			default:

				delete(entry);
				return PBO_STS_UNKNOWN_COMPRESSION;
			}
        }

		if (entry->hdr.reserved)
		{
			delete(entry);
			return PBO_STS_RESERVED_NONZERO;
		}
		switch(entry->hdr.type)
		{
		case ENTRYTYPE_COMPRESSED:
		
			entry->compressed_file=true;
			break;

		case ENTRYTYPE_NORMAL:
		
			if (!entry->compressed_file) break; // else sizes are different
		
		default:

			delete(entry);
			return PBO_STS_UNKNOWN_COMPRESSION;
			break;
		}

        if (!FirstEntry) FirstEntry = entry;
		if (tail) tail->next=entry;
		tail=entry;
        file_count++;
        entry->offset = file_size;
        file_size += entry->hdr.size;
        real_file_size += entry->hdr.real_size;
    }
}

int pboContainer::GetData(pbo_entry *entry)
{
int len= entry->ReadRawData(fpin, data_offset);
	if (len<=0) return status=len;
	// check that it's not squashed
	if (entry->data[0]) return len;
	if (strncmp((char *)&entry->data[1],"raP",3)) return len;
	bincpp.SetData((uchar*)entry->data,len);
char *fn=entry->hdr.filename;
char *newbuf=bincpp.ReadPackets();
	bincpp.Nullify();
	bincpp.Close();
	if (!newbuf) return len;
	free(entry->data);
	entry->data=(uchar*)newbuf;
	entry->extracted_file=true;
	return entry->actual_size=strlen(newbuf)+1;
}
int pboContainer::IsCompressed(pbo_entry *entry)
{
	return entry->compressed_file?1:0;
}

int pboContainer::WasEncrypted(pbo_entry *entry)
{
	return entry->extracted_file?1:0;
}

int pboContainer::GetData(char *Filename)
{
pbo_entry *entry=FindEntry(Filename);

	if (entry) return GetData(entry);
	return PBO_STS_NO_ENTRY;
}
pbo_entry * pboContainer::FindEntry(char *fname)
{
pbo_entry *entry=FirstEntry;

	while (entry)
	{
		if (!stricmp(entry->hdr.filename,fname))
			return entry;
		entry=entry->next;
	}
	return entry;
}
void pboContainer::FreeData(pbo_entry * entry)
{
	if (entry->data) free(entry->data);
	entry->data=NULL;
}
const char *pboContainer::strError()
{
	return strError(status);
}
const char *pboContainer::strError(int sts)
{
	switch(sts)
	{
	default: return "pbo Unknown error";
	case PBO_STS_NO_ERROR:	return "DePbo No error";
	case PBO_STS_NO_MEMORY:	return "DePbo No memory";
	case PBO_STS_FILE_TRUNCATED: return "DePbo File Truncated";
	case PBO_STS_NO_FILE: return "DePbo No File opened";
	case PBO_STS_OUT_OF_BOUNDS: return "DePbo compression out of bounds";
	case PBO_STS_CHUNK_ZERO: return "DePbo compression chunk is zero";
	case PBO_STS_CHECKSUM: return "DePbo compression checksum";
	case PBO_STS_NO_ENTRY: return "DePbo no entry found";
	case PBO_STS_MULIPLE_PRODUCTS: return "DePbo multiple product entries";
	case PBO_STS_UNKNOWN_COMPRESSION: return "DePbo unknown compression method";
	case PBO_STS_NON_ZERO_FIELD: return "DePbo an entry field expected to be zero";
	case PBO_STS_RESERVED_NONZERO: return "DePbo Entry's reserved field non-zero";
	}
}
