#include "gpl.h"
#include "stdafx.h"
#include "DePbo.h"
#include "Reader.h"
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include "ErrTypes.h"
#include "PboUtils.h"

class Reader Read;

PRIVATE bool Reader::AllocHold(const char *v)
{
char **newhold=(char **)realloc(HoldLine,(deflok+2)*sizeof(char **));

	if (!newhold) {DeAllocAllHolds(); return false;}
	HoldLine=newhold;
	HoldLine[deflok]=lineptr;
	HoldLine[deflok+1]=lineptr=strdup(v);
	if (!lineptr) return false;
	deflok+=2;
	return true;
}
PRIVATE void Reader::DeAllocHold()
{
	if (!deflok) return;
	deflok-=2;
	lineptr=HoldLine[deflok];
	free(HoldLine[deflok+1]);
	if (!deflok)
	{
		free(HoldLine);
		HoldLine=NULL;
	}
}
PRIVATE void Reader::DeAllocAllHolds()
{
	while (deflok) DeAllocHold();
}

Reader::Reader()
{
	Token=RawLine=NULL;
	FileBuffer=NULL;
	HoldLine=NULL;
	deflok=0;
	ProtectBufferName=false;
	RelaxedRules=true;// this for missing semicolon at eol
	Close();
}

Reader::~Reader(){Close();}
PUBLIC bool Reader::OpenFile(const char *sourcefile)
{
	Close();

FILE	*fpin;

	if ((fpin=fopen(sourcefile,"rb"))==NULL) return false;

	if (fseek(fpin,0,SEEK_END)) goto bad;
	
	FileCount=ftell(fpin);
	rewind(fpin);
	if (!(FilePtr=FileBuffer=( char *)malloc(FileCount))) goto bad;
	if (fread((char *)FileBuffer,1,FileCount,fpin)!=FileCount) 
	{
		Close();
bad:
		fclose(fpin);
		return false;
	}
	fclose(fpin);
	return true;
}
PUBLIC void Reader::SetBuffer(const char *BufferName,unsigned sz)
{
	Close();
	ProtectBufferName=true;
	FileBuffer=FilePtr=BufferName;
	FileCount=sz;
}


PUBLIC void Reader::Close()
{
	if (!ProtectBufferName && FileBuffer) free((char*)FileBuffer);
	FileBuffer=NULL;
	FileCount=0;
	asterdet=false;
	AllowMissingSemi=RelaxedRules;
	DeAllocAllHolds();
	lineptr=NULL;
	deflok=0;
	ProtectBufferName=false;
	if (RawLine) free(RawLine);
	if (Token) free(Token);
	Token=RawLine=NULL;
	TokenLen=RawLineLen=0;
}

PUBLIC int Reader::GetToken(int tt)
{
	while (deflok) 
	{
		if (*lineptr)	return ExpandIt(tt);
		DeAllocHold();
	}

	if (!lineptr || !*lineptr)
	{
		if (tt==GET_TOKEN_HUNT_EOL) return TokenReturn=TOKEN_EOL;
		switch(GetNextLine())
		{
		case 0: return ERR_NO_MEMORY; // or bad chars
		case EOF:
			if (AllowMissingSemi && tt==GET_TOKEN_TEST_EOL) return TokenReturn=';';

			return TokenReturn=TOKEN_EOF;
		}
		lineptr=RawLine;
		if (AllowMissingSemi && tt==GET_TOKEN_TEST_EOL)
		{
			return TokenReturn=';';
		}
	}
	if (tt==GET_TOKEN_HUNT_EOL)
	{
		strcpy(Token,lineptr);
		*lineptr=0;
		return TokenReturn=TOKEN_VALID;
	}
	return ExpandIt(tt);
}

PRIVATE int Reader::ExpandIt(int tt)
{
const char *ExpandedToken; // filled
int s;

	for (;;)
	{
		s= ProcessToken(tt);
		if (s!=TOKEN_VALID) return s;
		if (tt==GET_TOKEN_HUNT_DEF_LABEL) return s;
		if (tt==GET_TOKEN_ALLOW_SPACES) Trim(Token);
	/* now subsitute defines if relevent*/	
		switch(s=DefList->Find(Token,&ExpandedToken))
		{
		default: return s; // a parsing error occured
		case TOKEN_EOF: return TOKEN_VALID; // nothing there
		case TOKEN_VALID: break; // substitute has been performed
		}
		if (!AllocHold(ExpandedToken)) return ERR_NO_MEMORY;
	}
}

PRIVATE int Reader::ProcessToken(int tt)
{
bool brackdet=false;

char *TokPtr;

	if (!Token || TokenLen!=RawLineLen)
	{
		if (!(Token=(char*)malloc(TokenLen=RawLineLen))) return ERR_NO_MEMORY;
	}

	TokPtr=Token;

	*TokPtr=0;

	while (*lineptr)
	{
		switch(TokenReturn=*lineptr)
		{
		case '(':
				brackdet=true;
				goto ignore;
		case ')':

				*TokPtr++=*lineptr++;
				*TokPtr=0;
				return TokenReturn=TOKEN_VALID;

ignore:
		default: *TokPtr++=*lineptr++; break;
		case ' ':

			if (brackdet) goto ignore;
			if (tt==GET_TOKEN_ALLOW_SPACES)
			{
				if (TokPtr==Token) // suppress leading spaces and let trim knock the others
				{
					lineptr++;
					break;
				}
				goto ignore;
			}
			lineptr++;
			if (TokPtr!=Token)
			{
				*TokPtr=0; return TokenReturn=TOKEN_VALID;
			}
			break;

		case ':':			if (tt==GET_TOKEN_ALLOW_SPACES) goto ignore;
		case ',':
// semicolons at beginning of line are stripped and don't occur here
		case '=':
		case '{':
		case '}':
		case '[':
		case ']':
					if (tt==GET_TOKEN_ALLOW_SPACES) goto ignore;
		case ';':
			
			if (brackdet) goto ignore;
			if (TokPtr!=Token)
			{
				*TokPtr=0; return TokenReturn=TOKEN_VALID;
				// dont update linptr and it will come in clean
			}
			lineptr++;
			return TokenReturn;

		case '"':// in fact, these only belong as values so GET_TOKEN_ALLOW_SPACES
			// will always be true when a quote is detected

			if (brackdet) goto ignore;// probably not needed now

			if (TokPtr!=Token)
			{
				if (tt==GET_TOKEN_ALLOW_SPACES) goto ignore;
				*TokPtr=0; return TokenReturn=TOKEN_VALID;
			}
			/* have had issues here
			what i do know is hunt thru the entire quote and watch for matching ones
			then, at end, if only alpahmumercs exist, i drop the ""
				//	condition="this animationPhase ""Dvere1"" >= 0.5";
				//statement="this animate [""Dvere1"", 0]";
*/
			*TokPtr++=*lineptr++;
			while (*TokPtr++=*lineptr)
			{
				if (*lineptr++=='"')
				{
					if (*lineptr=='"') lineptr++; // a double?
					else
					{
						TokPtr[-1]=0;
							strcpy(Token,&Token[1]);
						return TokenReturn=TOKEN_VALID;
					}
				}
			}
			return TokenReturn=TOKEN_ERR_MISSING_QUOTE;
		}
	}
		// checkit's gone empty 
	if (TokPtr==Token) 	return TokenReturn=TOKEN_ERR_NO_TOKEN;

	*TokPtr=0;
	if (brackdet)return TokenReturn=ERR_CLOSEPAR_EXPECTED;
	return TokenReturn=TOKEN_VALID;
}
int Reader::FGETS()
{
	if (!RawLine)
	{
		if (!(RawLine=(char *)malloc(RawLineLen=100))) return 0; //fail
	}

char *end;
int len;

	if (!FileCount)		return EOF;

	end=strchr(FilePtr,'\n');
	if (!end) len=strlen(FilePtr)+1;
	else
	{
		len=end-FilePtr+1;
	}
	while (len>=RawLineLen)
	{
		if (!(RawLine=(char *)realloc(RawLine,RawLineLen*=2))) return 0; //fail
	}

char *line=RawLine;

	while (FileCount--)
	{
		if (!*FilePtr) return 0;
		if ((*line++=*FilePtr++)=='\n') 
		{
			*line=0;
			return 1;
		}
	}
	FileCount=0;
	*line=0;
	return 1;
}
PRIVATE int Reader::GetNextLine()
{
// alters \r\n to null
// converts any tab to a space

	// will ignore blank lines including whitespace
	//				lines with a semicolon in first column
	// strips any // as comment
bool stuff=false;
char *ptr;
int sts;

	while (!stuff)	// blank lines
	{
again:
		do
		{
			switch(FGETS())
			{
			case EOF: return EOF;
			case 0: return 0; // bugger
			default:
			case 1: break;
			}
			printf("%s",RawLine);
			ptr=RawLine;
			if (asterdet) 
			{
				for(ptr=RawLine;*ptr;)
				{
					if (*ptr++=='*')
					{
						if (*ptr=='/')
						{
							strcpy(RawLine,++ptr);
							asterdet=false;
							switch(*ptr)
							{
							case 0:
							case '\r':
							case '\n': goto again;
							}
							goto strip;
						}
					}
				}
				goto again;
			}
		} while (RawLine[0]==';'); // skip comments
// might also need to fix this for spaced ;
strip:
		ptr=RawLine;
		do
		{
			ptr=strchr(ptr,'/');	// strip ending comments

			if (ptr)
			{
				if (ptr[1]=='/') 
				{
					*ptr=0;
					if (!RawLine[0]) goto again;
					break;
				}
				if (ptr[1]=='*')
				{
					*ptr=0;
					ptr+=2;
					asterdet=true;

					for (;;)
					{
						switch(*ptr++)
						{
						case 0: goto aster_lineEnd;
												
						case '*':

							if (*ptr=='/')
							{
								strcat(RawLine,++ptr); // if anything
								asterdet=false;
								goto strip;
							}

						default: break;
						}
					}
				}
				ptr+=2;
			}
		}while (ptr);

// strip crlf, convert tabs to spaces, and check for non space
aster_lineEnd:
				/* comments remain in force*/
		for (ptr=RawLine;*ptr;ptr++)
		{
			switch(*ptr)
			{
			case '\r':
			case '\n': *ptr=0; goto done; 
			case '\t': *ptr=' '; 
			case ' ': break;
			default: stuff=true;
			}
		}
done: ;

	};
	Trim(RawLine);
	// concats
	/* very annoying but some sequences forget the ;
	*/
char *CatEnd=&RawLine[strlen(RawLine)-1];
	if (*CatEnd=='\\')
	{
		*CatEnd=0;
		char *hold;
		if (!(hold=strdup(RawLine))) return 0;
		if ((sts=GetNextLine())!=1)
		{
			free(hold);
			return sts;
		}
		int len=strlen(hold)+1;
		len+=strlen(RawLine);
		while (len>=RawLineLen)
		{
			if (!(RawLine=(char *)realloc(RawLine,RawLineLen*=2)))
			{
				free(hold);
				return 0;
			}
		}
		char *h2=strdup(RawLine);
		if (!h2) return 0;
		Trim(hold);
		if (hold[strlen(hold)-1]=='}')// bugger
				sprintf(RawLine,"%s;%s",hold,h2);
		else
				sprintf(RawLine,"%s %s",hold,h2);

		free(hold); free(h2);
	}
	return 1;
}
