/*
 * PBO file extractor
 * ------------------
 * This EXAMPLE program uses the dll to expand all files into their 
 * respective directories.
 *
 * syntax: MikePbo fullyqualified.pbo
 *
 *  will extract to fullyqualified dir
 *  returns 0 on success
 *
 */
#include "..\gpl.h"				// license agreement
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <direct.h>
#include <ctype.h>

#include "..\DePbo\DePbo.h"

#define mkdir(a) _mkdir(a)

#define VERSION "1.01"

static int makefile(char *dir,pbo_entry *entry,pboContainer *Container);

int main(int argc, char *argv[])
{
char *strDir, *ptr, *snap;
pbo_entry *entry;
pboContainer *Container;

    printf("ExtractPBO %s by mikero@norfolk.nf\n", VERSION);

    if (argc < 2)
    {
		ptr=strrchr(argv[0],'\\');
		if (ptr) ptr++; else ptr=argv[0];
        printf("Syntax: %s <pbo-file>\n", ptr);
        return 1;
    }
  
    if (!(snap=strDir = (char*)malloc(strlen(argv[1]) + 1)))
	{
		fprintf(stderr,"no memory");
        return 2;
	}
    ptr=strcpy(strDir, argv[1]);

    if ((ptr = strrchr(strDir, '/')))    strDir = ++ptr; // snap to actual fname

    if ( (!(ptr = strrchr(strDir, '.'))) || strlen(ptr)!=4 || stricmp(ptr,".pbo"))
	{
        fprintf(stderr, ".pbo file not declared (%s)\n",strDir);
		goto FreeSnapMemory;
    }
	*ptr = 0; // StrDir now contains the location of extraction directory (ie sans .pbo)

	if (!(Container=new pboContainer()))
	{
		printf("Nomem\n");
		goto FreeSnapMemory;
	}
	if (Container->Open(argv[1]))	// non zero = some error
    {
        fprintf(stderr,"%s\n",Container->strError());
		goto FreeSnapMemory;
	}

    mkdir(strDir);	// might already exist

	for (entry = Container->FirstEntry;entry;entry = Container->GetNextEntry(entry))
	{
		if (!makefile(strDir,entry,Container)) 
		{
			Container->Close();
FreeSnapMemory:
			free(snap);
			return 3;
		}
	}

	Container->Close();
	free(snap);
    return 0;
}

static int makefile(char *dir,pbo_entry *entry,pboContainer *Container)
{
FILE *fpout;
const char *EntryName=Container->GetEntryFilename(entry);
int len;
char *ptr,*fname;

	if (!EntryName || !*EntryName) return 0;
// whatever the name or folder is, it gets attached to the directory
	fname = (char*)malloc(strlen(dir) + strlen(EntryName) + 2);
	sprintf(fname,"%s/%s",dir,EntryName);
    /* Convert Backslashes into front, if any */
    for (ptr = fname;*ptr;ptr++)  	if (*ptr == '\\')		*ptr = '/';
	ptr=fname;
	do
	{
		ptr = strchr(ptr, '/');
		if (ptr)
		{
			*ptr = 0;
			mkdir(fname);
			*ptr++ = '/';
		}
    }while (ptr && *ptr);
	
	printf("%s\n",fname); // tell the world
    
	fpout = fopen(fname, "wb");
	free(fname);

	if (!fpout)
    {
		perror("Unable to open output file");
		return 0;
    }

    if ((len=Container->GetData(entry))<0)
	{
		fprintf(stderr,"%s\n",Container->strError()); goto cleanup;
	}
    fwrite(Container->GetDataPtr(entry), len, 1, fpout);
    fclose(fpout);
	Container->FreeData(entry);
	return 1;
cleanup:
	fclose(fpout);
	Container->FreeData(entry);
	return 0;
}
